/*
Copyright  2008 Brent Boyer

This program is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the Lesser GNU General Public License for more details.

You should have received a copy of the Lesser GNU General Public License along with this program (see the license directory in this project).  If not, see <http://www.gnu.org/licenses/>.
*/


import java.util.ArrayList;
import java.util.HashMap;
import java.util.TreeMap;
import java.util.Vector;
import java.util.concurrent.ConcurrentHashMap;

import bb.science.Lfsr;
import bb.util.Benchmark;


/**
* For the article
* <a href="http://www.ibm.com/developerworks/library/j-benchmark2/index.html">Robust Java benchmarking, Part 2: Statistics And solutions</a>,
* contains the code relevant to
* <a href="http://www.ibm.com/developerworks/library/j-benchmark2/index.html#fig3">Figure 3</a>,
* <a href="http://www.ibm.com/developerworks/library/j-benchmark2/index.html#fig4">Figure 4</a>,
* <a href="http://www.ibm.com/developerworks/library/j-benchmark2/index.html#fig5">Figure 5</a>,
* <a href="http://www.ibm.com/developerworks/library/j-benchmark2/index.html#listing2">Listing 2</a>,
* and <a href="http://www.ibm.com/developerworks/library/j-benchmark2/index.html#listing3">Listing 3</a>.
* <p>
* @author Brent Boyer
*/
public class N08_Part2_DataStructureAccess {
	
	
	// -------------------- static fields --------------------
	
	
	/**
	* Specifies the prefixes of the class names of all the task classes that will be benchmarked by {@link #main main}.
	* The classes must all be inner classes of this class.
	* The suffix of their class name must be specified as the 2nd elements of the args parameter of main.
	*/
	protected static String[] classnamePrefixes = new String[] {
		"ArrayAccess", "ArrayAccessSynchronized",	// array tasks
		"ArrayListAccess", "VectorAccess",	// List tasks
		"HashMapAccess", "TreeMapAccess", "ConcurrentHashMapAccess1", "ConcurrentHashMapAccess2",	// Map tasks
		"ThreadLocalAccess"	// ThreadLocal tasks
// +++ need to add Set tasks
// +++ add enum tasks (EnumSet and EnumMap)
	};
	
	
	/**
	* Global variable used by the various data structure access task classes like {@link ArrayAccessSequential}.
	* <p>
	* Is declared as a field here instead of as a field/local variable in the classes below because
	* <ol>
	*  <li>it is used by all of the other tasks below</li>
	*  <li>can change it in one single place and have it affect everything else</li>
	*  <li>it will never be garbage collected (unless reassigned)</li>
	* </ol>
	* Values are guaranteed to be unique Integer instances (integers[i].intValue() == i).
	*/
	protected static Integer[] integers;
	
	
	// -------------------- main and helper methods --------------------
	
	
	/**
	* Results on 2008/3/27 on my 2.2 GHz Intel Core 2 Duo E4500 machine (server jvm):
	* see these files:
	* <pre><code>
		../results/N08_Part2_DataStructureAccess_Desktop_1024_Sequential.txt
		../results/N08_Part2_DataStructureAccess_Desktop_1024_Random.txt
		../results/N08_Part2_DataStructureAccess_Desktop_1024X1024_Sequential.txt
		../results/N08_Part2_DataStructureAccess_Desktop_1024X1024_Random.txt
	</code></pre>
	* <p>
	* Results on 2008/3/27 on my 1.167 GHz SPARC-Enterprise-T5220 machine (server jvm):
	* see these files:
	* <pre><code>
		../results/N08_Part2_DataStructureAccess_N2_1024_Sequential.txt
		../results/N08_Part2_DataStructureAccess_N2_1024X1024_Sequential.txt
	</code></pre>
	* Note: the N2 results are CPU times, not elapsed times (since the tests are all single threaded and Solaris supports this so well)
	* <p>
	* The files mentioned above are the full text of the raw program output.
	* Edited versions of the relevant data (i.e. just the means) of all of the above results,
	* plus all of the associated Figures in the article, are found in this spreadsheet:
	* <pre><code>
		../results/N08_Part2_DataStructureAccess.xls
	</code></pre>
	* <p>
	* @param args must be non-null and have exactly 2 elements:
	* args[0] must be an int that specifies the size of the integers array,
	* args[1] must be either the String <code>Sequential</code> or the String <code>Random</code>
	* (it specifies how the elements of the integer array will be accessed)
	*/
	public static void main(String[] args) throws Exception {
		if (args == null) throw new IllegalArgumentException("args == null");
		if (args.length != 2) throw new IllegalArgumentException("args.length = " + args.length + " != 2");
		
			// parse m and carry out all consequences:
		int m = Integer.parseInt(args[0]);
		
		integers = new Integer[m];
		for (int i = 0; i < integers.length; i++) integers[i] = new Integer(i);
		
		Benchmark.Params params = new Benchmark.Params(true, m);	// true means measure many executions, m defines number of actions per execution
		//params.setMeasureCpuTime(true);	// default is false: measure elapsed time
		
		RandomIndex.init(m);
		
			// parse mode:
		String mode;
		if ("Sequential".equalsIgnoreCase(args[1])) mode = "Sequential";
		else if ("Random".equalsIgnoreCase(args[1])) mode = "Random";
		else throw new IllegalArgumentException("args[1] = " + args[1] + " !equals (case irrelevant) Sequential or Random");
		
		//examineConcurrentHashMapAnomaly(mode);	// uncomment this and the following line if just want to examine the ConcurrentHashMap anomaly
		//if (true) return;
		
		System.out.println();
		System.out.println("N08_Part2_DataStructureAccess (m = " + m + "):");
		for (String prefix : classnamePrefixes) {
			Runnable task = createTask(prefix, mode);
			System.out.println("\n" + prefix + mode + ": " + new Benchmark(task, params).toStringFull());
		}
	}
	
	
	@SuppressWarnings("unchecked")
	protected static Runnable createTask(String prefix, String mode) throws Exception {
		String classname = "N08_Part2_DataStructureAccess$" + prefix + mode;
		Class<Runnable> c = (Class<Runnable>) Class.forName(classname);	
		return c.newInstance();
	}
	
	
	/**
	* This method was written to more thoroughly examine the ConcurrentHashMap case 1 versus case 2 anomaly
	* noted in the article as a function of many more element sizes.
	* <p>
	* Results on 2008/5/2 on my 2.2 GHz Intel Core 2 Duo E4500 machine (server jvm):
	* <pre><code>
		m	case 1	case 2
		1	29.798 ns	32.366 ns
		2	28.773 ns	30.201 ns
		4	28.266 ns	25.859 ns
		8	25.361 ns	23.530 ns
		16	27.614 ns	24.478 ns
		32	30.415 ns	26.797 ns
		64	28.893 ns	26.729 ns
		128	29.057 ns	27.718 ns
		256	27.707 ns	27.575 ns
		512	28.846 ns	27.002 ns
		1024	31.326 ns	28.998 ns
		2048	34.486 ns	33.871 ns
		4096	33.757 ns	32.865 ns
		8192	35.414 ns	35.299 ns
		16384	35.290 ns	34.653 ns
		32768	36.722 ns	35.705 ns
		65536	71.930 ns	65.732 ns
		131072	124.900 ns	116.707 ns
		262144	158.120 ns	137.651 ns
		524288	174.524 ns	159.670 ns
		1048576	182.828 ns	164.059 ns
		2097152	194.498 ns	174.439 ns
		4194304	200.627 ns	172.409 ns
		8388608	211.306 ns	184.544 ns
	</code></pre>
* The above shows that the anomaly is always present, regardless of the size of the number of elements, which differs from what I found on an earlier run...
+++ try another run...
	*/
	protected static void examineConcurrentHashMapAnomaly(String mode) throws Exception {
		System.out.println();
		System.out.println("m" + "\t" + "case 1" + "\t" + "case 2");
		for (int m = 1; m <= 32 * 1024 * 1024; m *= 2) {
			integers = new Integer[m];
			for (int i = 0; i < integers.length; i++) integers[i] = new Integer(i);
			
			Benchmark.Params params = new Benchmark.Params(true, m);	// true means measure many executions, m defines number of actions per execution
			//params.setMeasureCpuTime(true);	// default is false: measure elapsed time
			
			Runnable task1 = createTask("ConcurrentHashMapAccess1", mode);
			double chm1 = new Benchmark(task1, params).getMean();
			Runnable task2 = createTask("ConcurrentHashMapAccess2", mode);
			double chm2 = new Benchmark(task2, params).getMean();
			System.out.println(m + "\t" + chm1 + "\t" + chm2);
		}
	}
	
	
	// -------------------- TaskRunnableAbstract --------------------
	
	
	// see http://www.ibm.com/developerworks/library/j-benchmark2/index.html#dsat for additional comments on the classes below
	
	
	/**
	* Abstract base class for all of the other Task classes.
	* The only functionality defined here is the scaffolding for DCE prevention in a Runnable task class.
	*/
	protected static abstract class TaskRunnableAbstract implements Runnable {
		protected int state = 0;	// needed to prevent DCE since this is a Runnable
		@Override public String toString() { return String.valueOf(state); }	// needed to prevent DCE since this is a Runnable
	}
	
	
	// -------------------- Sequential access versions of the task classes --------------------
	
	
	/**
	* Task class for benchmarking the access time of an array ({@link #integers}).
	* The elements are accessed in sequential order.
	*/
	protected static class ArrayAccessSequential extends TaskRunnableAbstract {
		public void run() {
			//for (int element : integers) {	// do NOT use this even for arrays: I benchmarked that it slows down the m = 1024 case by 5.1% (altho the m = 1024 * 1024 case was only slightly slower, the difference not beeing statistically significant)
			for (int i = 0; i < integers.length; i++) {
				state ^= integers[i];	// must use the data structure access to help compute the result in order to prevent DCE; bitwise XOR should be as fast an operation as there is, so it should less impact on the result
			}
		}
	}
	
	
	/** Same code as {@link ArrayAccessSequential} except synchronize on {@link #integers} around each access. */
	protected static class ArrayAccessSynchronizedSequential extends TaskRunnableAbstract {
		public void run() {
			for (int i = 0; i < integers.length; i++) {
				synchronized(integers) {
					state ^= integers[i];	// must use the data structure access to help compute the result in order to prevent DCE; bitwise XOR should be as fast an operation as there is, so it should less impact on the result
				}
			}
		}
	}
	
	
	/** Same code as {@link ArrayAccessSequential} except use an ArrayList that is equivalent to {@link #integers}. */
	protected static class ArrayListAccessSequential extends TaskRunnableAbstract {
		protected final ArrayList<Integer> arrayList = new ArrayList<Integer>(integers.length);
		{ for (Integer integer : integers) { arrayList.add(integer); } }
		
		public void run() {
			for (int i = 0; i < arrayList.size(); i++) {
				state ^= arrayList.get(i);	// must use the data structure access to help compute the result in order to prevent DCE; bitwise XOR should be as fast an operation as there is, so it should less impact on the result
			}
		}
	}
	
	
	/** Same code as {@link ArrayListAccessSequential} except use a Vector. */
	protected static class VectorAccessSequential extends TaskRunnableAbstract {
		protected final Vector<Integer> vector = new Vector<Integer>(integers.length);
		{ for (Integer integer : integers) { vector.add(integer); } }
		
		public void run() {
			for (int i = 0; i < vector.size(); i++) {
				state ^= vector.get(i);	// must use the data structure access to help compute the result in order to prevent DCE; bitwise XOR should be as fast an operation as there is, so it should less impact on the result
			}
		}
	}
	
	
	/** Same code as {@link ArrayAccessSequential} except use a HashMap that maps {@link #integers} to {@link #integers}. */
	protected static class HashMapAccessSequential extends TaskRunnableAbstract {
		protected final HashMap<Integer,Integer> hashMap = new HashMap<Integer,Integer>(integers.length);
		{ for (Integer integer : integers) { hashMap.put(integer, integer); } }
		
		public void run() {
			for (int i = 0; i < integers.length; i++) {
				state ^= hashMap.get(integers[i]);	// must use the data structure access to help compute the result in order to prevent DCE; bitwise XOR should be as fast an operation as there is, so it should less impact on the result
			}
		}
	}
	
	
	/** Same code as {@link HashMapAccessSequential} except use a TreeMap. */
	protected static class TreeMapAccessSequential extends TaskRunnableAbstract {
		protected final TreeMap<Integer,Integer> treeMap = new TreeMap<Integer,Integer>();
		{ for (Integer integer : integers) { treeMap.put(integer, integer); } }
		
		public void run() {
			for (int i = 0; i < integers.length; i++) {
				state ^= treeMap.get(integers[i]);	// must use the data structure access to help compute the result in order to prevent DCE; bitwise XOR should be as fast an operation as there is, so it should less impact on the result
			}
		}
	}
	
	
	/**
	* Same code as {@link HashMapAccessSequential} except use a ConcurrentHashMap.
	* This is a best case for ConcurrentHashMap access, since it specifies concurrencyLevel = 1.
	*/
	protected static class ConcurrentHashMapAccess1Sequential extends TaskRunnableAbstract {
		protected final ConcurrentHashMap<Integer,Integer> concMap = new ConcurrentHashMap<Integer,Integer>(integers.length, 0.75f, 1);
		{ for (Integer integer : integers) { concMap.put(integer, integer); } }
		
		public void run() {
			for (int i = 0; i < integers.length; i++) {
				state ^= concMap.get(integers[i]);	// must use the data structure access to help compute the result in order to prevent DCE; bitwise XOR should be as fast an operation as there is, so it should less impact on the result
			}
		}
	}
	
	
	/**
	* Same code as {@link HashMapAccessSequential} except use a ConcurrentHashMap.
	* This is a more realistic case for ConcurrentHashMap access, since it uses the default concurrencyLevel = 16.
	*/
	protected static class ConcurrentHashMapAccess2Sequential extends TaskRunnableAbstract {
		protected final ConcurrentHashMap<Integer,Integer> concMap = new ConcurrentHashMap<Integer,Integer>(integers.length);
		{ for (Integer integer : integers) { concMap.put(integer, integer); } }
		
		public void run() {
			for (int i = 0; i < integers.length; i++) {
				state ^= concMap.get(integers[i]);	// must use the data structure access to help compute the result in order to prevent DCE; bitwise XOR should be as fast an operation as there is, so it should less impact on the result
			}
		}
	}
	
	
	/** Similar code as {@link HashMapAccessSequential} except use a ThreadLocal. */
	@SuppressWarnings("unchecked")
	protected static class ThreadLocalAccessSequential extends TaskRunnableAbstract {
		protected final int m = integers.length;
		protected final ThreadLocal<Integer>[] integersPerThread = new ThreadLocal[m];
		{
			for (int i = 0; i < m; i++) {
				final int ii = i;
				integersPerThread[i] = new ThreadLocal<Integer>() {
					protected synchronized Integer initialValue() { return integers[ii]; }
				};
			}
		}
		
		public void run() {
			for (int i = 0; i < integersPerThread.length; i++) {
				state ^= integersPerThread[i].get();	// must use the data structure access to help compute the result in order to prevent DCE; bitwise XOR should be as fast an operation as there is, so it should less impact on the result
			}
		}
	}
	
	
	// -------------------- Random access versions of the task classes --------------------
	
	
	/**
	* Task class for benchmarking the access time of an array ({@link #integers}).
	* The elements are accessed in seemingly random order.
	*/
	protected static class ArrayAccessRandom extends ArrayAccessSequential {
		public void run() {
			for (int i = 0; i < integers.length; i++) {
				state ^= integers[RandomIndex.next()];	// must use the data structure access to help compute the result in order to prevent DCE; bitwise XOR should be as fast an operation as there is, so it should less impact on the result
			}
		}
	}
	
	
	/** Same code as {@link ArrayAccessRandom} except synchronize on {@link #integers} around each access. */
	protected static class ArrayAccessSynchronizedRandom extends ArrayAccessSynchronizedSequential {
		public void run() {
			for (int i = 0; i < integers.length; i++) {
				synchronized(integers) {
					state ^= integers[RandomIndex.next()];	// must use the data structure access to help compute the result in order to prevent DCE; bitwise XOR should be as fast an operation as there is, so it should less impact on the result
				}
			}
		}
	}
	
	
	/** Same code as {@link ArrayAccessRandom} except use an ArrayList that is equivalent to {@link #integers}. */
	protected static class ArrayListAccessRandom extends ArrayListAccessSequential {
		public void run() {
			for (int i = 0; i < arrayList.size(); i++) {
				state ^= arrayList.get(RandomIndex.next());	// must use the data structure access to help compute the result in order to prevent DCE; bitwise XOR should be as fast an operation as there is, so it should less impact on the result
			}
		}
	}
	
	
	/** Same code as {@link ArrayListAccessRandom} except use a Vector. */
	protected static class VectorAccessRandom extends VectorAccessSequential {
		public void run() {
			for (int i = 0; i < vector.size(); i++) {
				state ^= vector.get(RandomIndex.next());	// must use the data structure access to help compute the result in order to prevent DCE; bitwise XOR should be as fast an operation as there is, so it should less impact on the result
			}
		}
	}
	
	
	/** Same code as {@link ArrayAccessRandom} except use a HashMap that maps {@link #integers} to {@link #integers}. */
	protected static class HashMapAccessRandom extends HashMapAccessSequential {
		public void run() {
			for (int i = 0; i < integers.length; i++) {
				state ^= hashMap.get(integers[RandomIndex.next()]);	// must use the data structure access to help compute the result in order to prevent DCE; bitwise XOR should be as fast an operation as there is, so it should less impact on the result
			}
		}
	}
	
	
	/** Same code as {@link HashMapAccessRandom} except use a TreeMap. */
	protected static class TreeMapAccessRandom extends TreeMapAccessSequential {
		public void run() {
			for (int i = 0; i < integers.length; i++) {
				state ^= treeMap.get(integers[RandomIndex.next()]);	// must use the data structure access to help compute the result in order to prevent DCE; bitwise XOR should be as fast an operation as there is, so it should less impact on the result
			}
		}
	}
	
	
	/**
	* Same code as {@link HashMapAccessRandom} except use a ConcurrentHashMap.
	* This is a best case for ConcurrentHashMap access, since it specifies concurrencyLevel = 1.
	*/
	protected static class ConcurrentHashMapAccess1Random extends ConcurrentHashMapAccess1Sequential {
		public void run() {
			for (int i = 0; i < integers.length; i++) {
				state ^= concMap.get(integers[RandomIndex.next()]);	// must use the data structure access to help compute the result in order to prevent DCE; bitwise XOR should be as fast an operation as there is, so it should less impact on the result
			}
		}
	}
	
	
	/**
	* Same code as {@link HashMapAccessRandom} except use a ConcurrentHashMap.
	* This is a more realistic case for ConcurrentHashMap access, since it uses the default concurrencyLevel = 16.
	*/
	protected static class ConcurrentHashMapAccess2Random extends ConcurrentHashMapAccess2Sequential {
		public void run() {
			for (int i = 0; i < integers.length; i++) {
				state ^= concMap.get(integers[RandomIndex.next()]);	// must use the data structure access to help compute the result in order to prevent DCE; bitwise XOR should be as fast an operation as there is, so it should less impact on the result
			}
		}
	}
	
	
	/** Similar code as {@link HashMapAccessRandom} except use a ThreadLocal. */
	protected static class ThreadLocalAccessRandom extends ThreadLocalAccessSequential {
		public void run() {
			for (int i = 0; i < integersPerThread.length; i++) {
				state ^= integersPerThread[RandomIndex.next()].get();	// must use the data structure access to help compute the result in order to prevent DCE; bitwise XOR should be as fast an operation as there is, so it should less impact on the result
			}
		}
	}
	
	
	// -------------------- RandomIndex (static inner class) --------------------
	
	
	/**
	* Via the {@link #next next} method, returns all the valid indices of {@link #integers} in seemingly random order.
	* <p>
	* The current implementation internally uses a maximal length 32-bit linear feedback shift register (LFSR)
	* to continuously cycle thru all the valid indices of {@link #integers}
	* (except 0: LFSR's cannot handle it without getting stuck at it forever).
	* The code here is an adaptation of the class {@link Lfsr}.
	*/
	protected static class RandomIndex {
		
		protected static int register = 1;	// stores the LFSR state
		protected static int mask = 0xffffffff;	// actually, this is not needed for n = 32, only smaller n, but retain in case change n
		protected static int taps = (1 << 31) | (1 << 30) | (1 << 28) | (1 << 0);	// i.e. 32 31 29 1
		protected static int integersMask;
		
		protected static void init(int m) throws IllegalArgumentException {
			if (m < 2) throw new IllegalArgumentException("m = " + m + " < 2");
			
			for (int pow2 = 2; pow2 > 0; pow2 *= 2) {
				if (m == pow2) {
					integersMask = pow2 - 1;	// puts all 1s in the appropriate lower bits, and all 0s in the high bits
					return;
				}
			}
			throw new IllegalArgumentException("m = " + m + " is not supported because it is not a power of 2");
		}
		
		protected static int next() {
			register = ((register >>> 1) ^ (-(register & 1) & taps)) & mask;	// register & 1 selects the low order bit, and do - to create an int that is all 1's if was 1, or all 0's if was 0; the rest is easy to understand
			return register & integersMask;	// mask register with integersMask to ensure that the result is a valid index for integers
		}
		
	}
	
	
}